/*
 *	Sherlock Gatherer -- Object Queue
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>

#include "gather.h"

uns qobj_count;
static list queue;

void
oqueue_init(void)
{
  DIR *d;
  struct dirent *de;
  byte path[256];
  byte *z;
  struct stat st;
  list del_list;
  struct qobj *o, *oo, *r;

  init_list(&queue);
  init_list(&del_list);
  strcpy(path, TMP_DIR "/");
  z = path + strlen(path);

  if (! (d = opendir(TMP_DIR)))		/* Read all directory items */
    die("Cannot open temp directory");
  while (de = readdir(d))
    {
      byte *n = de->d_name;

      strcpy(z, n);
      if (lstat(path, &st))
	{
	  log(L_ERROR "Cannot stat %s", path);
	  continue;
	}
      if (!S_ISREG(st.st_mode))
	continue;
      o = xmalloc(sizeof(struct qobj));
      strcpy(o->tf.name, path);
      if (strlen(n) == 9 && n[0] == 'c')
	{
	  log(L_INFO "Found queued object %s", n+1);
	  strcpy(o->tf.name, path);
	  qobj_enqueue(o);
	}
      else
	add_tail(&del_list, &o->n);
    }
  closedir(d);

  o = HEAD(queue);			/* Handle data files */
  while (oo = (void *) o->n.next)
    {
      struct odes *d = obj_load(o->tf.name);
      byte *v;

      if (v = find_aval(d, '@'))
	{
	  DO_FOR_ALL(r, del_list)
	    if (r->tf.name[TMP_DIR_LEN + 1] == 'd' && !strcmp(r->tf.name, v))
	      {
		log(L_INFO "%s: using data file %s", o->tf.name, r->tf.name);
		rem_node(&r->n);
		free(r);
		goto del;
	      }
	  log(L_ERROR "%s: data file %s missing, object lost", o->tf.name, r->tf.name);
	  rem_node(&o->n);
	  add_tail(&del_list, &o->n);
	}
    del:
      obj_free(d);
      o = oo;
    }

  o = HEAD(del_list);			/* Delete all unused temp files */
  while (oo = (void *) o->n.next)
    {
      if (unlink(o->tf.name))
	log(L_ERROR "unlink(%s): %m", o->tf.name);
      else
	log(L_INFO "Deleted %s", o->tf.name);
      free(o);
      o = oo;
    }
}

struct qobj *
qobj_new(void)
{
  struct qobj *o = xmalloc(sizeof(struct qobj));

  open_temp(&o->tf, TF_QUEUE_CONTROL);
  return o;
}

void
qobj_del(struct qobj *o)
{
  maybe_unlink(o->tf.name);
  free(o);
}

void
qobj_enqueue(struct qobj *o)
{
  add_tail(&queue, &o->n);
  qobj_count++;
}

void					/* Enqueue named object descriptor */
qobj_enqueue_name(byte *name)
{
  struct qobj *o = xmalloc(sizeof(struct qobj));

  strcpy(o->tf.name, name);
  o->tf.fh = -1;
  qobj_enqueue(o);
}

struct qobj *
qobj_dequeue(void)
{
  struct qobj *o = HEAD(queue);

  if (o->n.next)
    {
      qobj_count--;
      rem_node(&o->n);
      return o;
    }
  else
    return NULL;
}
