/*
 *	Sherlock File Downloader
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

/*
 *	Calling convention:
 *
 *	Command line: <URL> [<if-mod-since> [<use-old-protocol>]]
 *
 *	fd[1] = object header
 *	fd[2] = log file
 */

/* FIXME:
 *		- normalization of input name!
 *		- if-modified-since
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <lib/lib.h>
#include <lib/url.h>

static uns if_modified_since;

static void invalid(byte *, ...) NONRET;
static void syserr(void) NONRET;

static void
invalid(byte *msg, ...)
{
  va_list args;

  va_start(args, msg);
  putchar('-');
  vprintf(msg, args);
  putchar('\n');
  exit(1);
}

static void
syserr(void)
{
  invalid("%d %m", (errno == EPERM) ? 1403 : 1404);
}

static void
do_file(byte *name, struct stat *st)
{
  struct tempfile t;
  int fd;
  uns n, r;
  byte buf[4096];

  if ((fd = open(name, O_RDONLY)) < 0)
    syserr();
  open_temp(&t, TF_QUEUE_DATA);
  printf("@%s\n", t.name);
  printf("s%u\n", (uns) st->st_size);
  r = st->st_size;
  while (r)
    {
      n = (r > sizeof(buf)) ? sizeof(buf) : r;
      if (read(fd, buf, n) != n)
	syserr();
      if (write(t.fh, buf, n) != n)
	syserr();
      r -= n;
    }
  puts("+Downloaded OK");
}

static int
mergelink(byte *to, byte *from, byte *with)
{
  byte *zz = to + MAX_URL_SIZE - 1;
  byte *p;
  uns z;

  if (*with == '/')
    {
      *to++ = *with++;
      p = to;
    }
  else
    {
      strcpy(to, from);
      p = strrchr(to, '/') + 1;
    }
  while (z = *with)
    {
      if (z == '/')
	{
	  with++;
	  continue;
	}
      if (z == '.')
	{
	  if (with[1] == '/' || !with[1])
	    {
	      with++;
	      continue;
	    }
	  if (with[1] == '.' && (with[2] == '/' || !with[2]))
	    {
	      with += 2;
	      if (p == to)
		return 0;
	      p--;			/* Slash */
	      while (p > to && p[-1] != '/')
		p--;
	      continue;
	    }
	}
      while (*with && *with != '/')
	if (p >= zz)
	  return 0;
	else
	  *p++ = *with++;
      if (*with)
	*p++ = *with++;
    }
  *p = 0;
  return 1;
}

static void
do_dir(byte *path, byte *sep)
{
  DIR *d;
  struct dirent *e;
  struct stat st;
  byte xbuf[MAX_URL_SIZE], ybuf[MAX_URL_SIZE], lbuf[MAX_URL_SIZE];
  byte *name;
  int l;

  strcpy(xbuf, path);
  sep = xbuf + (sep - path) + 1;
  if (chdir(path))
    syserr();
  d = opendir(".");
  if (!d)
    syserr();
  puts("X0125 Is a directory");
  while (e = readdir(d))
    {
      if (!strcmp(e->d_name, ".") || !strcmp(e->d_name, ".."))
	continue;
      if (sep + strlen(e->d_name) + 1 >= xbuf + MAX_URL_SIZE)
	continue;			/* Would be too long! */
      strcpy(sep, e->d_name);
      if (lstat(xbuf, &st))
	continue;
      if (S_ISLNK(st.st_mode))
	{
	  if (stat(xbuf, &st))
	    continue;
	  if ((l = readlink(xbuf, lbuf, sizeof(lbuf))) <= 0)
	    continue;
	  if (!mergelink(ybuf, xbuf, lbuf))
	    continue;
	  name = ybuf;
	}
      else
	name = xbuf;
      if (S_ISDIR(st.st_mode))
	printf("Rfile://localhost%s/\n", name);
      else if (S_ISREG(st.st_mode))
	printf("Rfile://localhost%s\n", name);
    }
  closedir(d);
}

static int
check_norm(byte *p)
{
  byte b[MAX_URL_SIZE];

  if (!mergelink(b, "/", p))
    invalid("0127 Invalid path");
  if (strcmp(p, b))
    {
      puts("X0129 Name not normalized");
      printf("Rfile://localhost%s\n", b);
      return 0;
    }
  return 1;
}

static void
do_it(byte *path)
{
  struct stat st;
  byte *z;
  int slashflag;

  if (!check_norm(path))
    return;

  z = strrchr(path, '/');
  if (!z)
    invalid("0127 Invalid file name");
  slashflag = !z[1];

  if (stat(path, &st))
    syserr();
  printf("L%u\n", (int) st.st_mtime);
  if (S_ISREG(st.st_mode))
    {
      if (slashflag)
	invalid("0128 Not a directory");
      do_file(path, &st);
    }
  else if (S_ISDIR(st.st_mode))
    {
      if (slashflag)
	do_dir(path, z);
      else
	{
	  puts("X0129 Directory redirect");
	  printf("Rfile://localhost%s/\n", path);
	}
    }
  else
    invalid("1403 Invalid Object type");
}

static void
get_url(byte *url)
{
  struct url u;
  byte buf1[MAX_URL_SIZE], buf2[MAX_URL_SIZE], buf3[MAX_URL_SIZE];
  int err;

  if (err = url_canon_split(url, buf1, buf2, &u))
    goto er;
  if (u.protoid != URL_PROTO_FILE)
    invalid("0100 Invalid protocol");
  if (u.host && u.host[0] && strcmp(u.host, "localhost"))
    invalid("0124 Host name must be `localhost' for the FILE URL scheme");
  if (u.user)
    invalid("0101 Authentication not supported");
  if (err = url_pack(&u, buf3))
    goto er;
  printf("U%s\n", buf3);
  printf("D%u\n", (int) time(NULL));
  do_it(u.rest);
  return;

er:
  invalid("00%02d %s", err, url_error(err));
}

int
main(int argc, char **argv)
{
  initlog(argv[0]);
  if (argc > 4 || argc < 2)
    die("Invalid arguments. Usage: fileget <url> [<if-mod-since> [<use-old-version-flag>]]");
  if (argc >= 3)
    if_modified_since = atol(argv[2]);
  get_url(argv[1]);
  return 0;
}
