/*
 *	Sherlock Database Builder -- Reference Splitter
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#include <lib/string.h>
#include <lib/lib.h>
#include <lib/searchdb.h>
#include <lib/wordcmp.h>
#include <lib/fastbuf.h>

#define BUFSIZE 4096

static struct fastbuf *iwl, *inl, *orl, *owl, *osl, *onl;

static ulg
getword(struct fastbuf *f, byte *w)
{
  int i;

  do
    {
      if (f->bptr < f->bstop)
	i = *f->bptr++;
      else if ((i = bgetc_slow(f)) == EOF)
	return 0;
      *w++ = i;
    }
  while (i);
  return bgetl(f);
}

static ulg nref;

static ulg
copyref(struct fastbuf *f, ulg c)
{
  ulg p = btell(orl);
  ulg porig;

  while ((p & (FILE_PAGE_SIZE-1)) > FILE_PAGE_SIZE - sizeof(c))
    {
      bputc(orl, 0);			/* Align on sizeof(ulg) */
      p++;
    }
  porig = p;
  nref += c;
  bputl(orl, c);
  p += sizeof(c);
  while (c)
    {
      uns r = FILE_PAGE_SIZE - (p & (FILE_PAGE_SIZE - 1)); /* Free space in the page */
      uns m = r / sizeof(struct sdbitem);
      if (!m)
	{
	  while (r--)			/* Pad to end of page */
	    {
	      bputc(orl, 0);
	      p++;
	    }
	  continue;
	}
      if (m > c)
	m = c;
      bbcopy(f, orl, m * sizeof(struct sdbitem));
      p += m * sizeof(struct sdbitem);
      c -= m;
    }
  return porig;
}

static void
splitw(void)
{
  byte word[MAX_WORD_LEN];
  byte lastsig[8], sig[8];
  ulg c, p, q;
  ulg wc = 0;

  lastsig[0] = 0;
  while (c = getword(iwl, word))
    {
      wc++;
      p = copyref(iwl, c);
      extract_sig(word, sig);
      if (memcmp(lastsig, sig, sizeof(sig)))
	{
	  if (lastsig[0])
	    bputc(owl, 0);
	  memcpy(lastsig, sig, sizeof(sig));
	  q = btell(owl);
	  bwrite(osl, sig, sizeof(sig));
	  bputl(osl, q);
	}
      bputs(owl, word);
      bputc(owl, 0);
      bputl(owl, p);
    }
  bputc(owl, 0);
  log(L_INFO "%d words", wc);
}

static void
splitn(void)
{
  byte word[MAX_WORD_LEN];
  ulg c, p;
  ulg nwc = 0;

  while (c = getword(inl, word))
    {
      p = copyref(inl, c);
      nwc++;
      bputs(onl, word);
      bputc(onl, 0);
      bputl(onl, p);
    }
  bputl(onl, nwc);
  log(L_INFO "%d non-words", nwc);
}

int
main(int argc, char **argv)
{
  initlog(argv[0]);
  if (argc != 7)
    die("Usage: splitter <in-wordlist> <in-nonwordlist> <out-reflist> <out-wordlist> <out-siglist> <out-nwlist>");
  log(L_INFO "Splitting database...");
  iwl = bopen(argv[1], O_RDONLY, BUFSIZE);
  inl = bopen(argv[2], O_RDONLY, BUFSIZE);
  orl = bopen(argv[3], O_WRONLY | O_CREAT | O_TRUNC, BUFSIZE);
  owl = bopen(argv[4], O_WRONLY | O_CREAT | O_TRUNC, BUFSIZE);
  osl = bopen(argv[5], O_WRONLY | O_CREAT | O_TRUNC, BUFSIZE);
  onl = bopen(argv[6], O_WRONLY | O_CREAT | O_TRUNC, BUFSIZE);
  splitw();
  splitn();
  bclose(iwl);
  bclose(inl);
  bclose(orl);
  bclose(owl);
  bclose(osl);
  bclose(onl);
  log(L_INFO "%d references", nref);
  return 0;
}
