/*
 *	Sherlock Database Builder -- Building of Word Signature SB-Tree
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#include <lib/lib.h>
#include <lib/searchdb.h>
#include <lib/fastbuf.h>

#undef TREE_DEBUG

#ifdef TREE_DEBUG
#define DBG(x, args...) log(L_DEBUG x, ## args)
#else
#define DBG(x, args...)
#endif

static struct fastbuf *sigs;
static int tree;
static uns nsigs, cpos;

static inline void
load_key(struct sig *dest, uns id, uns end)
{
  if (id >= end)
    memset(dest, 0xff, sizeof(struct sig));
  else
    {
      if (id != cpos)
	{
	  bsetpos(sigs, id * sizeof(struct sig));
	  cpos = id;
	}
      bread(sigs, dest, sizeof(struct sig));
    }
}

static void
write_page(struct tree_page *p)
{
  if (write(tree, p, sizeof(*p)) != sizeof(*p))
    die("Write failed: %m");
}

static uns
grow_subtree(uns first, uns cnt)
{
  uns depth, level, pos, stop, key;
  uns width[16];
  struct tree_page page;

  if (first >= nsigs)
    die("Aaargh. Out of eyes, out of mind...");
  if (first + cnt > nsigs)
    cnt = nsigs - first;
  stop = first + cnt;

  depth = 0;
  width[0] = 0;
  while (width[depth] < cnt)
    {
      width[depth+1] = (KEYS_PER_PAGE+1) * width[depth] + KEYS_PER_PAGE;
      depth++;
    }
  DBG("grow_subtree(%d,%d): depth=%d, width=%d", first, cnt, depth, width[depth]);

  bzero(&page, sizeof(page));
  for(level=0; level<depth; level++)
    {
      pos = first;
      while (pos < stop)
	{
	  for(key=0; key < KEYS_PER_PAGE; key++)
	    {
	      pos += width[depth - level - 1];
	      load_key(&page.keys[key], pos++, stop);
	    }
	  pos += width[depth - level - 1] + 1; /* +1 is in parent pages */
	  write_page(&page);
	}
    }
  return depth;
}

static void
grow(void)
{
  struct tree_page root;
  uns skip, i, j;
  uns maxd = 0;

  bzero(&root, sizeof(root));
  root.aux = nsigs;
  skip = nsigs / (KEYS_PER_PAGE + 1);	/* ceil((nsigs - KPP)/(KPP+1)) */
  i = skip;
  for(j=0; j < KEYS_PER_PAGE; j++)
    {
      load_key(&root.keys[j], i, nsigs);
      i += skip + 1;
    }
  write_page(&root);
  i = 0;
  for(j=0; j <= KEYS_PER_PAGE; j++)
    {
      uns d;
      DBG("Subtree %d at %d", j, lseek(tree, 0, SEEK_CUR));
      d = grow_subtree(i, skip);
      if (d > maxd)
	maxd = d;
      i += skip + 1;
    }
  log(L_INFO "Max depth = %d", maxd + 1);
}

int
main(int argc, char **argv)
{
  initlog(argv[0]);
  if (argc != 3)
    die("Usage: wtree <siglist> <tree>");
  log(L_INFO "Building signature tree...");
  sigs = bopen(argv[1], O_RDONLY, 65536);
  bseek(sigs, 0, SEEK_END);
  nsigs = btell(sigs) / sizeof(struct sig);
  cpos = ~0;
  log(L_INFO "%d signatures", nsigs);
  if ((tree = open(argv[2], O_WRONLY | O_CREAT | O_TRUNC, 0666)) < 0)
    die("Unable to create `%s': %m", argv[2]);
  grow();
  return 0;
}
