/*
 *	Sherlock Database Builder -- CGI User Interface
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 *
 *	Motto: "User (n.): A four-letter word."
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <setjmp.h>

#include <lib/lib.h>
#include <lib/lists.h>
#include <lib/string.h>
#include <lib/url.h>

/* CONFIGURATION */

#define SERVER_ADMIN "mailto:mj@gts.cz"	/* Administrator of the server */
#define SEARCH_ENGINE_IP 0x7f000001	/* IP address of sherlockd */
#define SEARCH_ENGINE_PORT 8192		/* Port of sherlockd */
#define SCRIPT_NAME "/~mj/query.cgi"	/* Name of this script */
#define SEARCH_ENGINE_TIMEOUT 30
#define MAX_REPLY_OBJECTS 128		/* Maximal # of objects in single reply */
#define NUMBER_OF_PAGES 10		/* How many pages of hits do we supply */
#define HITS_PER_PAGE 10

/* Generic names used in the form */

static byte *search_type_names[] = { "n", "a", NULL };
#define ST_NORMAL 0
#define ST_ADVANCED 1

static byte *section_names[] = { "all", "1", NULL };

static byte *display_names[] = { "n", "t", "d", NULL };
#define DISP_NORMAL 0
#define DISP_TERSE 1
#define DISP_DETAIL 2

static byte *lang_names[] = { "en", "cz", NULL };

static byte *in_type, *in_sections, *in_display, *in_lang, *in_query; /* Form parameters */
static uns search_type, search_section, display_mode, lang, page; /* Integer forms */
static uns obj_from, obj_to;

/* Rest of the form for differenct languages */

struct langinfo {
  byte *title, *head, **types, **sections, **dispmodes, **langnames, **cat, **server_msgs;
  uns server_msg_count;
};

static byte en_title[] = "The Sherlock Search Engine";
#define en_header en_title
static byte *en_search_types[] = { "Standard", "Advanced" };
static byte *en_section_names[] = { "all known data", "Czech Web" };
static byte *en_display_modes[] = { "standard", "terse", "detailed" };
static byte *en_lang_names[] = { "English", "Czech" };
static byte *en_cat[] = {
  "",
  " search through",
  " and display the results in",
  " form in",
  " language.",
  "Search for:",
  "Submit",
  "<P>If you don't know how to formulate queries, a complete <A HREF=\"/enhelp.html\">help page</A> is available.\n\
<P>If you have any questions or suggestions about how to improve Sherlock, contact \
<A HREF=\"" SERVER_ADMIN "\">the author</A>.\n",
  "<P>Here should be a hint text, but it isn't.\n",
  "<P>Invalid query string. Try again.\n",
  "<P>Translated to ",
  "Unable to connect to the search server",
  "Timed out when waiting for search server reply. Please try again.",
  "Server reply is incomplete. Please try again.",
  "Hits ",
  " to ",
  " of ",
  " matching documents.",
  "No matching documents."
};
static byte *en_serv_msgs[] = {
};

static struct langinfo en_lang = {
  en_title, en_header, en_search_types, en_section_names, en_display_modes,
  en_lang_names, en_cat, en_serv_msgs, sizeof(en_serv_msgs) / sizeof(byte *)
};

static byte cz_title[] = "Sherlock";
#define cz_header cz_title
static byte *cz_search_types[] = { "Obyejn", "Podrobn" };
static byte *cz_section_names[] = { "cel database", "eskho Webu" };
static byte *cz_display_modes[] = { "standardn", "zkrcen", "podrobn" };
static byte *cz_lang_names[] = { "anglicky", "esky" };
static byte *cz_cat[] = {
  "",
  " dotaz na prohledn",
  ", vsledky zobrazit v ",
  " form, a to ",
  ".",
  "Dotaz:",
  "Hledat",
  "<P>Pokud nevte, jak formulovat dotazy, je k dispozici <A HREF=\"/czhelp.html\">kompletn help</A>.\n\
<P>Mte-li jakkoliv otzky i npady, jak Sherlocka vylepit, obrate se \
<A HREF=\"" SERVER_ADMIN "\">na autora</A>.\n",
  "<P>Tady by ml bt hint text, ale zatm nen.\n",
  "<P>Chybn poloen dotaz - zkuste znovu.\n",
  "<P>Peloeno na ",			/* 10 */
  "Nemohu se spojit s vyhledvacm serverem.",
  "Vyhledvac server neodpovdl. Prosm, zkuste znovu.",
  "Odpov serveru je nekompletn. Prosm, zkuste znovu.",
  "Dokumenty ",
  " a ",
  " z ",
  " vyhovujcch.",
  "Dotazu nevyhovuj dn dokumenty"
};
static byte *cz_serv_msgs[] = {
  "101 Dotaz je pli dlouh",
  "102 Chybn syntaxe dotazu",
  "103 V dotazu uvedeno pli mnoho slov",
  "104 Na tento dotaz by bylo mono odpovdt vce mn libovolnm dokumentem",
  "105 Zadanmu dotazu vyhovuje pli mnoho slov"
};

static struct langinfo cz_lang = {
  cz_title, cz_header, cz_search_types, cz_section_names, cz_display_modes,
  cz_lang_names, cz_cat, cz_serv_msgs, sizeof(cz_serv_msgs) / sizeof(byte *)
};

static struct langinfo *languages[] = { &en_lang, &cz_lang };
static struct langinfo *L;

static void xerr(byte *msg, ...) NONRET;

static void xerr(byte *msg, ...)
{
  va_list args;

  va_start(args, msg);
  puts("Content-type: text/html\n");
  puts("<HTML><HEAD><TITLE>Internal error</TITLE><BODY>\n<H1>Internal error</H1><P>");
  vfprintf(stdout, msg, args);
  puts("\n<P>Please <A HREF=\"" SERVER_ADMIN "\">notify server administrator</A>");
  puts("</BODY></HTML>");
  exit(0);
}

static void
parse_env(void)
{
  byte *g = getenv("REQUEST_METHOD");
  byte *q = getenv("QUERY_STRING");
  byte buf[MAX_URL_SIZE];
  byte *z, c, *p;

  if (!g || !q || strcmp(g, "GET"))
    xerr("CGI passed invalid environment");
  if (strlen(q) > MAX_URL_SIZE - 2)
    xerr("You passed an URL exceeding the maximum supported length");

  while (*q)
    {
      if (*q == '&')
	{
	  q++;
	  continue;
	}
      g = q;
      z = NULL;
      while (*q && *q != '&')
	if (*q == '+')
	  *q++ = ' ';
	else if (*q == '=')
	  z = q++;
	else
	  q++;
      if (!z)
	continue;
      *z++ = 0;
      c = *q;
      *q = 0;
      if (!url_deescape(z, buf))
	{
	  z = stralloc(buf);
	  for(p=z; *p; p++)
	    if (*p < NCC_MAX)
	      *p = NCC_CHARS[*p];
	  if (!strcmp(g, "type"))
	    in_type = z;
	  else if (!strcmp(g, "sec"))
	    in_sections = z;
	  else if (!strcmp(g, "disp"))
	    in_display = z;
	  else if (!strcmp(g, "lang"))
	    in_lang = z;
	  else if (!strcmp(g, "q"))
	    in_query = z;
	  else if (!strcmp(g, "pg"))
	    page = atol(z);
	}
      *q = c;
    }
}

static uns
decode(byte *s, byte **x)
{
  uns i = 0;

  if (!s)
    return 0;
  while (*x)
    {
      if (!strcasecmp(s, *x))
	return i;
      x++, i++;
    }
  return 0;
}

static void
decode_env(void)
{
  search_type = decode(in_type, search_type_names);
  search_section = decode(in_sections, section_names);
  display_mode = decode(in_display, display_names);
  lang = decode(in_lang, lang_names);
  L = languages[lang];
  if (page <= 0 || page > NUMBER_OF_PAGES)
    page = 1;
  obj_from = (page-1) * HITS_PER_PAGE;
  obj_to = obj_from + HITS_PER_PAGE - 1;
}

static void
choices(byte **vals, byte **names, uns is)
{
  uns i = 0;
  while (*vals)
    printf("<OPTION VALUE=%s%s>%s", *vals++, (i++ == is) ? " SELECTED" : "", *names++);
  printf("</SELECT>");
}

static void
send_escaped(byte *x)
{
  for(; *x; x++)
    if (*x == '&')
      printf("&amp;");
    else if (*x == '"')
      printf("&quot;");
    else if (*x == '<')
      printf("&lt;");
    else
      putchar(*x);
}

static void
header(void)
{
  byte **c = L->cat;

  puts("Content-type: text/html\n");
  printf("<!DOCTYPE HTML PUBLIC \"-//W3C/DTD HTML 3.1//EN\">\n");
  printf("<HTML><HEAD><TITLE>%s</TITLE></HEAD><BODY>\n", L->title);
  printf("<H1>%s</H1>\n", L->head);
  printf("<FORM METHOD=GET ACTION=\"" SCRIPT_NAME "\">\n");
  printf("<P>%s<SELECT NAME=type>", c[0]);
  choices(search_type_names, L->types, search_type);
  printf("%s\n<SELECT NAME=sec>", c[1]);
  choices(section_names, L->sections, search_section);
  printf("%s\n<SELECT NAME=disp>", c[2]);
  choices(display_names, L->dispmodes, display_mode);
  printf("%s\n<SELECT NAME=lang>", c[3]);
  choices(lang_names, L->langnames, lang);
  printf("%s\n", c[4]);
  printf("<P>%s <INPUT NAME=q SIZE=60 MAXLENGTH=240 VALUE=\"", c[5]);
  send_escaped(in_query ? in_query : (byte *) "");
  printf("\">\n<INPUT TYPE=SUBMIT VALUE=%s></FORM>\n\n", c[6]);
}

static void
footer(void)
{
  printf("<HR>%s", L->cat[7]);
  printf("</BODY></HTML>\n");
}

static byte *
xform_query(byte *q)
{
  byte *w[32], *t;
  uns i, m, j, k, zf;
  byte *dest = xmalloc(MAX_URL_SIZE + 1024);

  i = m = 0;
  q = stralloc(q);
  while (*q)
    {
      if (Cspace(*q))
	q++;
      else if (i < sizeof(w) / sizeof(byte *))
	{
	  t = w[i++] = q;
	  while (*q && !Cspace(*q))
	    if (*q++ == '"')
	      return NULL;
	  if (*q)
	    *q++ = 0;
	  if (t[0] == '+')
	    {
	      m++;
	      if (!t[1])
		return NULL;
	    }
	}
      else
	return NULL;
    }
  if (!i)
    return NULL;

  t = dest;
  zf = 0;
  if (m)
    {
      k = i;
      for(j=0; j<i; j++)
	if (w[j][0] == '+')
	  {
	    t += sprintf(t, "\"%s\"", w[j]+1);
	    if (--k)
	      t += sprintf(t, " AND ");
	  }
      zf = 1;
    }
  if (i != m)
    {
      k = i - m;
      if (zf)
	t += sprintf(t, "(ANY OR ");
      for(j=0; j<i; j++)
	if (w[j][0] != '+')
	  {
	    t += sprintf(t, "\"%s\"", w[j]);
	    if (--k)
	      t += sprintf(t, " OR ");
	  }
      if (zf)
	*t++ = ')';
    }
  *t = 0;
  return dest;
}

struct line {
  struct line *next;
  byte t[1];
};

static struct line *hdrs, *objs[MAX_REPLY_OBJECTS], *finish;
static uns nobjs;
static jmp_buf timeout_j;
static FILE *infile;

static void
timed_out(int unused)
{
  longjmp(timeout_j, 12);
}

static struct line *
read_em_all(void)
{
  byte buf[4096];
  byte *e;
  struct line *l, **ll, *q;

  l = NULL;
  ll = &l;

  for(;;)
    {
      if (!fgets(buf, sizeof(buf), infile))
	longjmp(timeout_j, 13);
      e = strchr(buf, '\n');
      if (!e)
	longjmp(timeout_j, 13);
      *e = 0;
      if (!buf[0])
	break;
      q = xmalloc(sizeof(struct line) + (e - buf));
      *ll = q;
      ll = &q->next;
      q->next = NULL;
      strcpy(q->t, buf);
    }
  return l;
}

static byte *
find_str(struct line *l, uns i)
{
  while (l)
    {
      if (l->t[0] == i)
	return l->t + 1;
      l = l->next;
    }
  return NULL;
}

static uns
find_num(struct line *l, uns i, uns def)
{
  byte *r = find_str(l, i);

  return r ? atol(r) : def;
}

static int
ask_the_server(byte *q, byte **em)
{
  int sock, l, e;
  struct sockaddr_in sa;
  byte *buf = alloca(strlen(q) + 256);
  byte *r;

  sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
  if (sock < 0)
    xerr("Unable to create socket: %m");

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = htonl(SEARCH_ENGINE_IP);
  sa.sin_port = htons(SEARCH_ENGINE_PORT);
  if (connect(sock, (struct sockaddr *) &sa, sizeof(sa)))
    return 11;

  l = sprintf(buf, "FROM %d TO %d %s%s%s\r\n",
	      obj_from,
	      obj_to,
	      search_section ? " SECTION " : "",
	      search_section ? (char *) search_section : "",
	      q);
  if (write(sock, buf, l) != l)
    return 12;

  if (!(infile = fdopen(sock, "r")))
    xerr("fdopen failed: %m");

  if (!(e = setjmp(timeout_j)))
    {
      signal(SIGALRM, timed_out);
      alarm(SEARCH_ENGINE_TIMEOUT);
      hdrs = read_em_all();
      if (r = find_str(hdrs, '-'))
	{
	  *em = r;
	  return -1;
	}
      nobjs = find_num(hdrs, 'N', 0);
      if (nobjs > MAX_REPLY_OBJECTS)
	xerr("nobjs .GT. MAX_REPLY_OBJECTS");
      for(l=0; l<nobjs; l++)
	objs[l] = read_em_all();
      finish = read_em_all();
    }
  else
    return e;

  return 0;
}

static byte *
lookup_server_msg(byte *m)
{
  byte **table = L->server_msgs;
  uns cnt = L->server_msg_count;

  if (strlen(m) < 4)
    return m;
  while (cnt--)
    {
      if (!memcmp(m, *table, 4))
	return *table + 4;
      table++;
    }
  return m+4;
}

static void
display_objs(void)
{
  uns i, j, k, max;
  byte *b, *u;
  struct line *l;
  time_t t;

  printf("<P>%s%d%s%d%s%d%s\n",
	 L->cat[14],
	 obj_from + 1,
	 L->cat[15],
	 obj_from + nobjs,
	 L->cat[16],
	 find_num(hdrs, 'T', 0),
	 L->cat[17]);

  if (display_mode >= DISP_DETAIL)
    {
      printf("<P><TT>\n");
      for(l=hdrs; l; l=l->next)
	if (l->t[0] == 'w')
	  {
	    sscanf(b = l->t+1, "%d", &j);
	    while (*b && *b != '\t')
	      b++;
	    if (*b)
	      b++;
	    printf("%s %d<BR>\n", b, j);
	  }
    }

  printf("<P><OL START=%d>\n", obj_from+1);
  for(i=0; i<nobjs; i++)
    {
      struct line *o = objs[i];
      printf("<LI>");
      u = find_str(o, 'U');
      if (!u)
	u = "?? NO URL ??";
      if (b = find_str(o, 'H'))
	{
	  printf("<A HREF=\"%s\">", u);
	  send_escaped(b);
	  puts("</A><BR>");
	}

      max = (display_mode == DISP_TERSE) ? !b : ~0;
      j = 0;
      for(l=o; l; l=l->next)
	if (l->t[0] == 'U' && max)
	  {
	    max--;
	    u = l->t + 1;
	    printf("<A HREF=\"%s\">", u);
	    send_escaped(u);
	    printf("</A>");
	    if (!j && display_mode == DISP_NORMAL)
	      {
		j = 1;
		printf(" (");
		if (t = find_num(o, 'L', 0))
		  {
		    struct tm *tm;
		    if (tm = localtime(&t))
		      printf("%02d-%02d-%02d, ", tm->tm_mday, tm->tm_mon+1, tm->tm_year%100);
		  }
		if ((k = find_num(o, 's', ~0)) != ~0)
		  printf("%u bytes, ", k);
		printf("q=%d)", find_num(o, 'Q', 0));
	      }
	    puts("<BR>");
	  }

      if (display_mode == DISP_DETAIL)
	{
	  if ((k = find_num(o, 's', ~0)) != ~0)
	    printf("Size: %u, ", k);
	  if (t = find_num(o, 'L', 0))
	    {
	      struct tm *tm;
	      if (tm = localtime(&t))
		printf("Date: %02d-%02d-%02d, ", tm->tm_mday, tm->tm_mon+1, tm->tm_year%100);
	    }
	  if (b = find_str(o, 'c'))
	    printf("Charset: %s, ", b);
	  printf("Q: %d<BR>\n", find_num(o, 'Q', 0));
	}

      if (display_mode != DISP_TERSE && (b = find_str(o, 'C')))
	{
	  send_escaped(b);
	  puts("");
	}
    }
  puts("</OL>");
}

static void
disp_bot(uns page, byte *z)
{
  byte u[MAX_URL_SIZE];
  byte r[8];
  byte *a, *b;

  if (url_enescape(in_query, u))
    strcpy(u, "???");
  a = b = u;
  while (*a)
    if (*a == '%' && a[1] == '2' && a[2] == '0')
      a+=3, *b++ = '+';
    else
      *b++ = *a++;
  *b = 0;
  if (!z)
    {
      sprintf(r, "%d", page);
      z = r;
    }
  printf("\n<A HREF=\"" SCRIPT_NAME "?type=%s&sec=%s&disp=%s&lang=%s&pg=%d&q=%s\">%s</A>",
	 search_type_names[search_type],
	 section_names[search_section],
	 display_names[display_mode],
	 lang_names[lang],
	 page,
	 u,
	 z);
}

static void
display_bottom(void)
{
  uns i;
  uns lap = (find_num(hdrs, 'T', 1) + HITS_PER_PAGE - 1) / HITS_PER_PAGE;

  if (lap == 1)
    return;
  if (lap > NUMBER_OF_PAGES)
    lap = NUMBER_OF_PAGES;
  printf("<P ALIGN=CENTER>");
  if (page > 1)
    disp_bot(page - 1, "&lt;");
  for(i=1; i<=lap; i++)
    if (i == page)
      printf(" %d", i);
    else
      disp_bot(i, NULL);
  if (page < lap)
    disp_bot(page + 1, "&gt;");
  puts("");
}

static void
do_query(void)
{
  byte *q, *em;
  int e;

  if (search_type == ST_NORMAL)
    {
      q = xform_query(in_query);
      if (!q)
	{
	  printf("<HR>%s\n", L->cat[9]);
	  return;
	}
    }
  else
    q = in_query;

  printf("<HR>");

  if (display_mode >= DISP_DETAIL)
    {
      printf("<P>%s ", L->cat[10]);
      send_escaped(q);
      putchar('\n');
    }

  if (e = ask_the_server(q, &em))
    {
      byte *m = (e < 0) ? lookup_server_msg(em) : L->cat[e];
      printf("<P><EM>%s</EM>\n", m);
      return;
    }

  if (nobjs)
    {
      display_objs();
      display_bottom();
    }
  else
    printf("<P>%s\n", L->cat[18]);
}

static void
do_hint(void)
{
  printf("<HR>%s\n", L->cat[8]);
}

int
main(int argc, char **argv)
{
  if (argc > 1)
    xerr("Invalid arguments");
  parse_env();
  decode_env();
  header();
  if (in_query && in_query[0])
    do_query();
  else
    do_hint();
  footer();
  return 0;
}
