/*
 *	Sherlock Search Engine -- Configuration File
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <fcntl.h>
#include <unistd.h>

#include "sherlock.h"

#include <lib/string.h>

static byte *log_file;

byte *obj_pool_name, *obj_idx_name, *ref_list_name, *sig_tree_name;
byte *word_list_name, *nword_list_name, *nword_hash_name;
uns port = 4444, listen_queue = 4, log_incoming, log_rejected, log_requests;
uns num_matches = 100, cache_size = 10, max_cache_age = 0, max_words = HARD_MAX_WORDS;
uns max_word_matches = HARD_MAX_WORD_MATCHES, file_cache_size = 16, found_word_bonus;
uns default_word_weight = 100, log_timing, connection_timeout = 60;

static list access_list;

struct access_entry {
  node n;
  uns allow;
  ulg addr, mask;
};

static byte *
parse_ip(byte *x, ulg *a)
{
  uns i, q;
  ulg z = 0;

  for(i=0; i<4; i++)
    {
      q = 0;
      while (Cdigit(*x))
	{
	  q = q*10 + *x++ - '0';
	  if (q > 255)
	    return "Invalid IP address";
	}
      if (*x++ != ((i == 3) ? 0 : '.'))
	return "Invalid IP address";
      z = (z << 8) | q;
    }
  *a = z;
  return NULL;
}

static byte *
cf_access(struct cfitem *c, byte *b)
{
  byte *p = strchr(b, '/');
  byte *q;
  struct access_entry *a = xmalloc(sizeof(struct access_entry));

  a->allow = (c->name[0] == 'A');
  if (p)
    {
      *p++ = 0;
      if (q = parse_ip(p, &a->mask))
	return q;
    }
  else
    a->mask = ~0;
  add_tail(&access_list, &a->n);
  return parse_ip(b, &a->addr);
}

int
check_access(ulg ip)
{
  struct access_entry *a;

  DO_FOR_ALL(a, access_list)
    if (! ((ip ^ a->addr) & a->mask))
      return a->allow;
  return 0;
}

static struct cfitem cfitems[] = {
  { "LogFile", CI_STRING, &log_file },
  { "ObjectPool", CI_STRING, &obj_pool_name },
  { "ObjectIndex", CI_STRING, &obj_idx_name },
  { "ReferenceList", CI_STRING, &ref_list_name },
  { "SignatureTree", CI_STRING, &sig_tree_name },
  { "WordList", CI_STRING, &word_list_name },
  { "NonWordList", CI_STRING, &nword_list_name },
  { "NonWordHash", CI_STRING, &nword_hash_name },
  { "Port", CI_INT, &port },
  { "ListenQueue", CI_INT, &listen_queue },
  { "Allow", CI_FUNCTION, cf_access },
  { "Deny", CI_FUNCTION, cf_access },
  { "LogIncoming", CI_INT, &log_incoming },
  { "LogRejected", CI_INT, &log_rejected },
  { "LogRequests", CI_INT, &log_requests },
  { "NumMatches", CI_INT, &num_matches },
  { "CacheSize", CI_INT, &cache_size },
  { "CacheAge", CI_INT, &max_cache_age },
  { "MaxWords", CI_INT, &max_words },
  { "MaxWordMatches", CI_INT, &max_word_matches },
  { "FileCachePages", CI_INT, &file_cache_size },
  { "WordBonus", CI_INT, &found_word_bonus },
  { "WordWeight", CI_INT, &default_word_weight },
  { "LogTiming", CI_INT, &log_timing },
  { "ConnTimeOut", CI_INT, &connection_timeout },
  { NULL, CI_STOP, NULL }
};

void
read_config(void)
{
  init_list(&access_list);
  cf_read("cf/sherlockd", cfitems);
  open_log_file(log_file);
}
