/*
 *	Sherlock Search Engine -- Searching for Non-Words
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "sherlock.h"

#include <lib/searchdb.h>
#include <lib/url.h>
#include <lib/pools.h>

static int hash_fd, nlist_fd;
static uns hash_val;

static inline uns
nw_hash(byte *w)
{
  uns i = 0;
  do
    {
      i = ((i * HASH_MAGIC) + *w) % hash_val;
    }
  while (*w++);
  return i;
}

int
lookup_nonword(byte *w, ulg *x)
{
  uns h = nw_hash(w);
  uns l, s;
  ulg *ppos, pos, ref;
  byte buf[MAX_WORD_LEN], *b, *r, *re;

  ppos = (ulg *) read_data(hash_fd, h * sizeof(ulg), &l);
  if (l < sizeof(ulg))
    die("lookup_nonword(): Lost sword... Awaiting doom...");
  pos = *ppos;
  DBG("%s: hash %d, pos %d", w, h, pos);
  b = buf;
  s = 0;
  for(;;)
    {
      r = read_data(nlist_fd, pos, &l);
      re = r + l;
      pos += l;
      while (r < re)
	{
	  if (s)
	    {
	      *b++ = *r++;
	      if (!--s)
		{
		  if (!strcmp(buf, w))
		    {
		      *x = ref;
		      return 1;
		    }
		  b = buf;
		}
	    }
	  else if (*r)
	    *b++ = *r++;
	  else
	    {
	      *b = *r++;
	      if (b == buf)
		return 0;
	      b = (byte *) &ref;
	      s = sizeof(ref);
	    }
	}
    }
  return 0;
}

void
nonword_init(void)
{
  hash_fd = open_file(nword_hash_name);
  nlist_fd = open_file(nword_list_name);
  hash_val = lseek(hash_fd, 0, SEEK_END) / sizeof(ulg);
  DBG("Non-word list: %d buckets", hash_val);
}

byte *
normalize_nonword(struct mempool *p, byte *a)
{
  byte b1[MAX_URL_SIZE], b2[MAX_URL_SIZE], b3[MAX_URL_SIZE], b4[MAX_URL_SIZE];
  int e;
  struct url url;

  if (   (e = url_canon_split(a, b1, b2, &url))
      || (e = url_pack(&url, b3))
      || (e = url_enescape(b3, b4)))
    {
      sprintf(b1, "-%d %s", 200+e, url_error(e));
      err(b1);
    }
  DBG("Normalized `%s' to `%s'", a, b4);
  a = pool_alloc(p, 1 + strlen(b4));
  strcpy(a, b4);
  return a;
}
