/*
 *	Sherlock Search Engine -- Query Parser
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>

#include "sherlock.h"

#include <lib/pools.h>
#include <lib/searchdb.h>

#define yyerror(x) err(x)

static byte *error;

static struct expr *new_node(uns t)
{
  struct expr *n = fast_alloc(current_query->pool, sizeof(struct expr));
  n->type = t;
  return n;
}

%}

%union {
  int i;
  struct query *q;
  struct expr *x;
  byte *s;
}

/* Tokens */

%token EOLN SECTION FROM TO ANY
%token <i> REF NUM
%token <s> STRING

%left OR
%left AND
%left NOT

%type <x> expr
%type <i> fromspec tospec refmask quality rlist

%%

main:
   section fromspec tospec expr EOLN {
      struct query *q = current_query;
      q->from = $2;
      q->to = $3;
      q->expr = $4;
      return 0;
   }
 ;

section:
   /* empty */ {
      memset(current_query->sections, 0xff, sizeof(current_query->sections));
   }
 | SECTION seclist
 ;

seclist:
   secno
 | seclist ',' secno
 ;

secno:
   NUM {
      current_query->sections[$1 >> 3] |= (1 << ($1 & 7));
   }
 ;

fromspec:
   /* empty */ {
     $$ = 0;
   }
 | FROM NUM {
     $$ = $2;
   }
 ;

tospec:
   /* empty */ {
     $$ = ~0;
   }
 | TO NUM {
     $$ = $2;
   }
 ;

expr:
   '(' expr ')' {
     $$ = $2;
   }
 | expr AND expr {
     $$ = new_node(EX_AND);
     $$->u.op.l = $1;
     $$->u.op.r = $3;
   }
 | expr OR expr {
     $$ = new_node(EX_OR);
     $$->u.op.l = $1;
     $$->u.op.r = $3;
   }
 | NOT expr {
     $$ = new_node(EX_NOT);
     $$->u.op.l = $2;
     $$->u.op.r = NULL;
   }
 | refmask quality STRING {
     $$ = new_node(EX_MATCH);
     $$->u.match.quality = $2;
     $$->u.match.classmap = $1 ? $1 : (1 << WT_WORD);
     $$->u.match.is_nonword = !! $1;
     $$->u.match.word = ($$->u.match.is_nonword ? normalize_nonword : normalize_word) (current_query->pool, $3);
   }
| ANY {
     $$ = new_node(EX_ANY);
   }
 ;

refmask:
   /* empty */ {
     $$ = 0;
   }
 | rlist
 ;

rlist:
   REF {
     $$ = $1 ? (1 << $1) : ~(1 << WT_WORD);
   }
 | rlist ',' REF {
     $$ = $1 | ($3 ? (1 << $3) : ~(1 << WT_WORD));
   }
 ;

quality:
   /* empty */ {
     $$ = default_word_weight;
   }
 | NUM {
     if ($1 < 0 || $1 > 1000)
       err("Quality out of range");
     $$ = $1;
   }
 ;

%%

static jmp_buf err_jmp;

void
err(byte *x)
{
  error = x;
  longjmp(err_jmp, 1);
}

byte *
parse_query(byte *b)
{
  error = NULL;
  lex_init(b);
  if (!setjmp(err_jmp))
    yyparse();
  return error;
}
