/*
 *	Sherlock Search Engine -- Declarations
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <lib/lib.h>
#include <lib/lists.h>

#ifdef BIG_DEBUG
#define DBG(x, y...) log(L_DEBUG x, ##y)
#else
#define DBG(x, y...)
#endif

#ifdef DETAIL_DEBUG			/* Define locally in the file being debugged... */
#define DDBG(x, y...) DBG(x, ##y)
#else
#define DDBG(x, y...)
#endif

/* config.c */

void read_config(void);
int check_access(ulg);

extern byte *obj_pool_name, *obj_idx_name, *ref_list_name, *sig_tree_name;
extern byte *word_list_name, *nword_list_name, *nword_hash_name;
extern uns port, listen_queue, log_incoming, log_rejected, log_requests;
extern uns num_matches, cache_size, max_cache_age, max_words, max_word_matches;
extern uns file_cache_size, found_word_bonus, default_word_weight, log_timing;
extern uns connection_timeout;

/* request.c */

#define BUF_SIZE 4096

struct reply {
  struct reply *next;
  uns len;
  byte text[1];
};

struct replybuf {
  struct reply *first, **last;
  struct mempool *pool;
};

struct query {
  node n;
  ulg established;			/* Time this connection was established */
  byte sections[256/8];			/* Section tag bitmap */
  uns from, to;				/* Result range selection */
  struct expr *expr;			/* Query expression */
  struct mempool *pool;			/* Memory pool */
  struct replybuf rbuf;			/* Reply header */
  int fd;
  byte outbuf[BUF_SIZE];		/* Output buffering */
  byte *obptr, *obend;
  byte inbuf[BUF_SIZE];
  byte *ibptr, *ibend;
  uns parse_msec, map_msec,		/* Timing */
      words_msec, refs_msec, objs_msec;
};

extern struct query *current_query;
extern struct replybuf *query_rbuf, *cache_rbuf;

struct query *new_query(void);
void free_query(struct query *);
void init_reply_buf(struct replybuf *, struct mempool *);
void add_reply(struct replybuf *, byte *, ...);
void prepend_reply(struct replybuf *, byte *, ...);
void reply_string(struct query *, byte *, uns);
void ship_reply_buf(struct query *, struct replybuf *);
void flush_reply_buf(struct query *, struct replybuf *);
void flush_query(struct query *);

/* lex.c */

void lex_init(byte *);
int yylex(void);

/* parse.y */

struct expr {
  byte type;
  union {
    struct {
      struct expr *l, *r;
    } op;
    struct {
      byte *word;
      word quality;
      byte classmap;
      byte is_nonword;
      byte wd_index;
    } match;
  } u;
};

#define EX_AND 0
#define EX_OR 1
#define EX_NOT 2
#define EX_MATCH 3
#define EX_ANY 4

void err(byte *) NONRET;
byte *parse_query(byte *);

/* query.c */

#define HARD_MAX_WORDS 16		/* Need memory for 2^this bits */
#define HARD_MAX_WORD_MATCHES 64

void normalize_query(struct query *, byte *);
void cache_init(void);
void do_query(struct query *);
void ref_init(void);

/* words.c */

int lookup_word(byte *, ulg *);
void word_init(void);
byte *normalize_word(struct mempool *, byte *);

/* nonwords.c */

int lookup_nonword(byte *, ulg *);
void nonword_init(void);
byte *normalize_nonword(struct mempool *, byte *);

/* files.c */

void file_cache_init(void);
int open_file(byte *);
byte *read_page(int, ulg);
byte *read_data(int, ulg, uns *);
void *lock_last_page(void);
void unlock_page(void *);

/* objects.c */

void objects_init(void);
void show_object(struct query *, ulg);
